//*****************************************************
//**   Ivory Draw 1.0 Beta 4                         **
//**   DirectDraw component package for C++Builder   **
//*****************************************************

// (C) 1998 Tamas Demjen
// email: demjen@usa.net
// Web: http://members.xoom.com/demjen
// Ivory Draw on the Web: http://members.xoom.com/demjen/bcb/ivory.html

// This product is freeware, as described in license.txt
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "CustomIvorySurface.h"
#include "IvoryDraw.h"
//---------------------------------------------------------------------------
#ifdef __IvoryBCB3__
#pragma package(smart_init)
#endif
//---------------------------------------------------------------------------
// internal variables:
static const TIvoryErrorClass iec = iecCustomIvorySurface;
//---------------------------------------------------------------------------
static inline void ValidCtrCheckCustomSurface(TCustomIvorySurface*)
{
   new TCustomIvorySurface(NULL);
}
//---------------------------------------------------------------------------
__fastcall TCustomIvorySurface::TCustomIvorySurface(TComponent* AOwner)
 : TIvoryResFileBase(AOwner),
   FDirectDrawSurface(NULL), FDirectDrawClipper(NULL), FCanvas(NULL),
   FAutoCreate(true), FSurfaceType(stUnspecified), FWidth(0), FHeight(0),
   FTransparent(false), FTransparentColor(0), FLocked(0), FPixels(NULL),
   FPitch(0), FClearBackground(true), FBackgroundColor(0), FPainted(false),
   FOnCreateInternal(NULL), FOnSurfaceLost(NULL), FOnRequestSize(NULL),
   FOnPaint(NULL), FOnCreate(NULL), FOnDestroy(NULL), FOnChange(NULL),
   FValid(false)
{
   FCanvas = new TIvoryCanvas(); // create a new canvas
   Canvas->OnChange = OnChange;
   if(TheIvoryDraw) { Width = TheIvoryDraw->Width; Height = TheIvoryDraw->Height; }
}
//---------------------------------------------------------------------------
__fastcall TCustomIvorySurface::~TCustomIvorySurface()
{
   if(DirectDrawSurface)
      TheIvoryDraw->ReleaseSurface(this);
}
//---------------------------------------------------------------------------
unsigned char* __fastcall TCustomIvorySurface::GetScanLine(int Row) const
{
   if(Pixels == NULL) return NULL;
   else return Pixels + Row * Pitch;
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::SetAutoCreate(bool ac)
{
   if(DirectDrawSurface == NULL) // AutoCreate can only be set before creating an off-screen surface
      FAutoCreate = ac;
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::SetWidth(int width)
{
   if(DirectDrawSurface == NULL && FWidth != width) // the width can only be set before creating an off-screen surface
   {
      FWidth = width;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::SetHeight(int height)
{
   if(DirectDrawSurface == NULL && FHeight != height) // the height can only be set before creating an off-screen surface
   {
      FHeight = height;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::SetTransparent(bool transparent)
{
   if(DirectDrawSurface == NULL && FTransparent != transparent) // the transparency can only be set before creating an off-screen surface
   {
      FTransparent = transparent;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::SetTransparentColor(int tcolor)
{
   if(DirectDrawSurface == NULL && FTransparentColor != tcolor) // the transparent color can only be set before creating an off-screen surface
   {
      FTransparentColor = tcolor;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::SetClearBackground(bool cl)
{
   if(FClearBackground != cl)
   {
      FClearBackground = cl;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::SetBackgroundColor(int bc)
{
   if(FBackgroundColor != bc)
   {
      FBackgroundColor = bc;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::SetLocked(bool locked)
{
   if(FLocked != locked)
   {
      if(locked == true)
         Lock();
      else
         Unlock();
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::CreatePrimary()
{
   const char* ivoryfuncname = "CreatePrimary";
   if(DirectDrawSurface == NULL)
   {
      Invalidate();
      HRESULT ddrval;
      FSurfaceType = stPrimary; // gonna be a primary surface
      DDSURFACEDESC ddsd;
      ZeroMemory(&ddsd, sizeof(ddsd));
      ddsd.dwSize = sizeof(ddsd);
      if(TheIvoryDraw->ExclusiveMode) // if we're in exclusive mode
      {
         ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
         ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
         ddsd.dwBackBufferCount = 1;
      }
      else // if we're not in exclusive mode
      {
         ddsd.dwFlags = DDSD_CAPS;
         ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
      }
      ddrval = TheIvoryDraw->DirectDraw->CreateSurface(&ddsd, &FDirectDrawSurface, NULL);
      if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t create surface", ierDirectX, ddrval);
      FWidth = TheIvoryDraw->Width; FHeight = TheIvoryDraw->Height;
      if(OnCreateInternal) OnCreateInternal(this);
      if(OnCreate) OnCreate(this);
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::CreateBack(TCustomIvorySurface* primary)
{
   const char* ivoryfuncname = "CreateBack";
   if(DirectDrawSurface == NULL)
   {
      Invalidate();
      HRESULT ddrval;
      if(TheIvoryDraw->ExclusiveMode) // if we're in exclusive mode
      {
         DDSCAPS ddscaps;
         FSurfaceType = stBack; // gonna be a back surface
         ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
         ddrval = primary->DirectDrawSurface->GetAttachedSurface(&ddscaps, &FDirectDrawSurface);
         if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t get attached surface", ierDirectX, ddrval);
      }
      else // if we're not in exclusive mode
      {
         HRESULT ddrval;
         DDSURFACEDESC ddsd;

         FSurfaceType = stOffScreen; // gonna be an off-screen surface
         ZeroMemory(&ddsd, sizeof(ddsd));
         ddsd.dwSize = sizeof(ddsd);
         ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
         ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
         ddsd.dwHeight = TheIvoryDraw->Height;
         ddsd.dwWidth = TheIvoryDraw->Width;
         ddrval = TheIvoryDraw->DirectDraw->CreateSurface(&ddsd, &FDirectDrawSurface, NULL);
         if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t create surface", ierDirectX, ddrval);

         ddrval = TheIvoryDraw->DirectDraw->CreateClipper(0, &FDirectDrawClipper, NULL);
         if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t create clipper", ierDirectX, ddrval);
         ddrval = DirectDrawClipper->SetHWnd(0, TheIvoryDraw->Handle);
         if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t set HWND for clipper", ierDirectX, ddrval);
         ddrval = primary->DirectDrawSurface->SetClipper(DirectDrawClipper);
         if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t set clipper", ierDirectX, ddrval);
      }
      FWidth = TheIvoryDraw->Width; FHeight = TheIvoryDraw->Height;
      if(OnCreateInternal) OnCreateInternal(this);
      if(OnCreate) OnCreate(this);
//      Paint();
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::InternalCreateOffScreen()
{
   const char* err_msg = "Couldn\'t create off-screen surface";
   const char* ivoryfuncname = "InternalCreateOffScreen";
   if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);
   if(DirectDrawSurface)
   {
      if(SurfaceType != stOffScreen) return; // primary and back surfaces cannot be recreated
//      BeforeDestruction();
      Release();
   }

   if(OnRequestSize) OnRequestSize(this, FWidth, FHeight);

   FSurfaceType = stOffScreen; // gonna be an off-screen surface
   Invalidate();

   HRESULT ddrval;
   DDSURFACEDESC ddsd;

   ZeroMemory(&ddsd, sizeof(ddsd));
   ddsd.dwSize = sizeof(ddsd);
   ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
   ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
   ddsd.dwHeight = Height;
   ddsd.dwWidth = Width;
   ddrval = TheIvoryDraw->DirectDraw->CreateSurface(&ddsd, &FDirectDrawSurface, NULL);
   if(ddrval != DD_OK) IVORYDDERROR(err_msg, ierDirectX, ddrval);
   TheIvoryDraw->AddSurface(this);
   if(Transparent) // if transparent, set the transparent color
   {
      DDCOLORKEY ddck;
      ddck.dwColorSpaceLowValue = TransparentColor;
      ddck.dwColorSpaceHighValue = TransparentColor;
      DirectDrawSurface->SetColorKey(DDCKEY_SRCBLT, &ddck);
   }

   if(OnCreateInternal) OnCreateInternal(this);
   if(OnCreate) OnCreate(this);
// Paint(); // paint the surface
}
//---------------------------------------------------------------------------
// release the Direct Draw Surface object (clear it from the memory)
void __fastcall TCustomIvorySurface::Release()
{
   if(OnDestroy) OnDestroy(this);
   if(Canvas->Handle) ReleaseCanvas();
   if(Locked) Unlock();
   if(DirectDrawClipper)
      DirectDrawClipper->Release(); // I'm not really sure if it needs here
   ReleaseCanvas(); // release the device context grabbed from the surface
   if(DirectDrawSurface)
   {
      if(SurfaceType != stBack) // if it's not a back surface
         DirectDrawSurface->Release(); // release the Direct Draw Surface
      FDirectDrawSurface = NULL;
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::RealizePaint()
{
   if(ClearBackground)
   {
      if(TheIvoryDraw->ExclusiveMode)
         FillRect(Rect(0, 0, Width, Height), BackgroundColor);
      else
      {
         GrabCanvas();
         TColor c = ToTColor(BackgroundColor);
         Canvas->Pen->Color = c;
         Canvas->Brush->Color = c;
         Canvas->Rectangle(0, 0, Width, Height);
      }
   }
   if(OnPaint) OnPaint(this);
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::StartPainting()
{
   if(TheIvoryDraw) TheIvoryDraw->StartPainting(this);
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::FinishPainting()
{
   if(TheIvoryDraw) TheIvoryDraw->FinishPainting(this);
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::Create()
{
   if(DirectDrawSurface == NULL)
   {
//      if(OnCreateInternal) OnCreateInternal(this);
      if(DirectDrawSurface == NULL && Width > 0 && Height > 0)
         InternalCreateOffScreen();
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::Recreate(int width, int height)
{
   if(Width == width && Height == height) return;
   FWidth = width;
   FHeight = height;
   if(Width > 0 && Height > 0)
      InternalCreateOffScreen();
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::Recreate(int width, int height, bool transparent, int color)
{
   if(Width == width && Height == height && Transparent == transparent && TransparentColor == color) return;
   FWidth = width;
   FHeight = height;
   FTransparent = transparent;
   FTransparentColor = color;
   if(Width > 0 && Height > 0)
      InternalCreateOffScreen();
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::Paint(bool subpaint)
{
   if(Valid) return;
   const char* err_msg = "Error painting surface";
   const char* ivoryfuncname = "Paint";
   if(subpaint)
   {
      if(FPainted) return;
      FPainted = true;
   }
   if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);
   if(DirectDrawSurface == NULL) IVORYERROR(err_msg, ierIvorySurfaceNotCreated);
   FSubPaint = subpaint;
   if(subpaint)
   {
      StartPainting();
      try
      {
         RealizePaint();
         Valid = true;
         if(Canvas->Handle) ReleaseCanvas();
         if(Locked) Unlock();
      }
      catch(...)
      {
         FinishPainting();
         throw;
      }
      FinishPainting();
   }
   else
   {
      RealizePaint();
      if(Canvas->Handle) ReleaseCanvas();
      if(Locked) Unlock();
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::Lock()
{
   const char* err_msg = "Couldn\'t lock surface";
   const char* ivoryfuncname = "Lock";
   if(!FLocked)
   {
      if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);
      if(DirectDrawSurface == NULL) IVORYERROR(err_msg, ierIvorySurfaceNotCreated);

      if(Canvas->Handle) ReleaseCanvas();
      DDSURFACEDESC ddsd;
      ZeroMemory(&ddsd, sizeof(ddsd));
      ddsd.dwSize = sizeof(ddsd);
      HRESULT ddrval;
//    RECT r = Rect(0, 0, Width, Height);
      ddrval = DirectDrawSurface->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
      if(ddrval != DD_OK) IVORYDDERROR(err_msg, ierDirectX, ddrval);
      FPixels = (unsigned char*)ddsd.lpSurface;
      FPitch = ddsd.lPitch;
      FLocked = true;
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::Unlock()
{
   const char* err_msg = "Couldn\'t unlock surface";
   const char* ivoryfuncname = "Unlock";
   if(FLocked)
   {
      if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);
      if(DirectDrawSurface == NULL) IVORYERROR(err_msg, ierIvorySurfaceNotCreated);
      HRESULT ddrval;
      ddrval = DirectDrawSurface->Unlock(NULL);
      if(ddrval != DD_OK) IVORYDDERROR(err_msg, ierDirectX, ddrval);
      FPixels = NULL;
      FPitch = 0;
      FLocked = false;
      Invalidate();
      if(OnChange) OnChange(this);
   }
}
//---------------------------------------------------------------------------
// get a device context from the surface
void __fastcall TCustomIvorySurface::GrabCanvas()
{
   if(Canvas->Handle == NULL)
   {
      const char* err_msg = "Couldn\'t grab canvas";
      const char* ivoryfuncname = "GrabCanvas";
      if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);
      if(DirectDrawSurface == NULL) IVORYERROR(err_msg, ierIvorySurfaceNotCreated);
      if(Locked) Unlock(); // automatically unlock the surface if necessary
      HDC hdc;
      HRESULT ddrval;
      ddrval = DirectDrawSurface->GetDC(&hdc); // get the context
      if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t get Device Context", ierDirectX, ddrval);
      Canvas->SetHandle(hdc); // pass the context to the canvas object
   }
}
//---------------------------------------------------------------------------
// release the device context grabbed from the surface
void __fastcall TCustomIvorySurface::ReleaseCanvas()
{
   if(Canvas->Handle)
   {
      const char* err_msg = "Couldn\'t release canvas";
      const char* ivoryfuncname = "ReleaseCanvas";
      if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);
      if(DirectDrawSurface == NULL) IVORYERROR(err_msg, ierIvorySurfaceNotCreated);
      DirectDrawSurface->ReleaseDC(Canvas->Handle); // release the context
      Canvas->SetHandle(NULL); // pass NULL to the canvas object
   }
}
//---------------------------------------------------------------------------
// Copy another surface
void __fastcall TCustomIvorySurface::CopySurface(int origx, int origy, TCustomIvorySurface* surface)
{
   CopySurface(origx, origy, surface, Rect(0, 0, surface->Width, surface->Height));
}
//---------------------------------------------------------------------------
// Copy a rectangle from another surface
void __fastcall TCustomIvorySurface::CopySurface(int origx, int origy, TCustomIvorySurface* surface, const TRect rect)
{
   const char* err_msg = "Couldn\'t copy surface";
   const char* ivoryfuncname = "CopySurface";
   if(surface == NULL || surface->DirectDrawSurface == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);
   int options;
   if(surface->Transparent) options = DDBLTFAST_SRCCOLORKEY;
   else options = DDBLTFAST_NOCOLORKEY;
   if(SurfaceType == stBack) options |= DDBLTFAST_WAIT;
   HRESULT ddrval;
   ddrval = DirectDrawSurface->BltFast(origx, origy, surface->DirectDrawSurface, (RECT*)&rect, options);
   if(ddrval != DD_OK) IVORYDDERROR(err_msg, ierDirectX, ddrval);
   Invalidate();
   if(OnChange) OnChange(this);
}
//---------------------------------------------------------------------------
// Stretch copy another surface
void __fastcall TCustomIvorySurface::StretchCopySurface(const TRect destrect, TCustomIvorySurface* surface)
{
   StretchCopySurface(destrect, surface, Rect(0, 0, surface->Width, surface->Height));
}
//---------------------------------------------------------------------------
// Stretch copy a rectangle from another surface
void __fastcall TCustomIvorySurface::StretchCopySurface(const TRect destrect, TCustomIvorySurface* surface, const TRect rect)
{
   const char* err_msg = "Couldn\'t stretch surface";
   const char* ivoryfuncname = "StretchCopySurface";
   if(surface == NULL || surface->DirectDrawSurface == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);
   int options;
   if(surface->Transparent) options = DDBLT_KEYSRC;
   else options = DDBLTFAST_NOCOLORKEY;
   if(SurfaceType == stBack) options |= DDBLT_WAIT;
   HRESULT ddrval;
   DDBLTFX ddbltfx;
   ZeroMemory(&ddbltfx, sizeof(ddbltfx));
   ddbltfx.dwSize = sizeof(ddbltfx);
   ddrval = DirectDrawSurface->Blt((RECT*)&destrect, surface->DirectDrawSurface, (RECT*)&rect, options, &ddbltfx);
   if(ddrval != DD_OK) IVORYDDERROR(err_msg, ierDirectX, ddrval);
   Invalidate();
   if(OnChange) OnChange(this);
}
//---------------------------------------------------------------------------
// Tile copy another surface
void __fastcall TCustomIvorySurface::TileCopySurface(const TRect destrect, TCustomIvorySurface* surface)
{
   TileCopySurface(destrect, surface, Rect(0, 0, surface->Width, surface->Height));
}
//---------------------------------------------------------------------------
// Tile copy a rectangle from another surface
void __fastcall TCustomIvorySurface::TileCopySurface(const TRect destrect, TCustomIvorySurface* surface, const TRect rect)
{
   const char* err_msg = "Couldn\'t tile surface";
   const char* ivoryfuncname = "TileCopySurface";
   if(surface == NULL || surface->DirectDrawSurface == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);

   int x = destrect.Left, y = destrect.Top, width = abs(destrect.Right-destrect.Left), height = abs(destrect.Bottom-destrect.Top);
   int srcwidth = abs(rect.Right-rect.Left), srcheight = abs(rect.Bottom-rect.Top);
   if(rect.Left == rect.Right || rect.Top == rect.Bottom)
      return;
   int horizN = width / srcwidth;
   int horizR = width - horizN * srcwidth;
   int vertN = height / srcheight;
   int vertR = height - vertN * srcheight;
   int currX, currY;

   for(currY = y; currY < vertN * srcheight; currY += srcheight)
   {
      for(currX = x; currX < horizN * srcwidth; currX += srcwidth)
         CopySurface(currX, currY, surface, rect);
      if(horizR)
         CopySurface(currX, currY, surface, Rect(0, 0, horizR, srcheight));
   }
   if(vertR)
   {
      for(currX = x; currX < horizN * srcwidth; currX += srcwidth)
         CopySurface(currX, currY, surface, Rect(0, 0, srcwidth, vertR));
      if(horizR)
         CopySurface(currX, currY, surface, Rect(0, 0, horizR, vertR));
   }
}
//---------------------------------------------------------------------------
void __fastcall TCustomIvorySurface::FillRect(const TRect rect, int color)
{
   const char* err_msg = "Couldn\'t fill rectangle";
   const char* ivoryfuncname = "FillRect";
   HRESULT ddrval;
   DDBLTFX ddbltfx;
   ZeroMemory(&ddbltfx, sizeof(ddbltfx));
   ddbltfx.dwSize = sizeof(ddbltfx);
   ddbltfx.dwFillColor = color;
   ddrval = DirectDrawSurface->Blt((RECT*)&rect, NULL, NULL, DDBLT_COLORFILL, &ddbltfx);
   if(ddrval != DD_OK) IVORYDDERROR(err_msg, ierDirectX, ddrval);
   Invalidate();
   if(OnChange) OnChange(this);
}
//---------------------------------------------------------------------------

