//*****************************************************
//**   Ivory Draw 1.0 Beta 4                         **
//**   DirectDraw component package for C++Builder   **
//*****************************************************

// (C) 1998 Tamas Demjen
// email: demjen@usa.net
// Web: http://members.xoom.com/demjen
// Ivory Draw on the Web: http://members.xoom.com/demjen/bcb/ivory.html

// This product is freeware, as described in license.txt
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "IvoryDraw.h"
#include "IvorySurface.h"
//---------------------------------------------------------------------------
#ifdef __IvoryBCB3__
#pragma package(smart_init)
USERES("ivory.dcr");
#endif
//---------------------------------------------------------------------------
// public variables:
TIvoryDraw* TheIvoryDraw = NULL;
   // Pointer to the Ivory Draw component. Only 1 Ivory Draw component can be
   // active at a time.
//---------------------------------------------------------------------------
// internal variables:
static TList* DDSList;
   // Direct Draw Surface list
static TList* DDPList;
   // Direct Draw Palette list
static HINSTANCE hDll = NULL;
   // DirectDraw DLL instance
typedef HRESULT __stdcall (*PROCDirectDrawCreate)
   (GUID FAR* lpGUID, LPDIRECTDRAW FAR* lplpDD, IUnknown FAR* pUnkOuter);
   // Pointer type to the DirectDrawCreate function
static PROCDirectDrawCreate PDirectDrawCreate = NULL;
   // Pointer to the DirectDrawCreate function
static LPDIRECTDRAW TheDirectDraw = NULL;
   // Pointer to the currently or lately activated TDirectDraw component, or if
   // you have never activated a TDirectDraw component, then to the lately
   // dropped component, or if you have never dropped a component, then it's NULL
static const TIvoryErrorClass iec = iecIvoryDraw;
//---------------------------------------------------------------------------
static inline void ValidCtrCheckDraw(TIvoryDraw*)
{
   new TIvoryDraw(NULL);
}
//---------------------------------------------------------------------------
TColor __fastcall ToTColor(int palindex)
{
   const char* err_msg = "Couldn\'t convert palette index to RGB color";
   const char* ivoryfuncname = "::ToTColor";
   if(TheIvoryDraw == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);
   if(TheIvoryDraw->Palette == NULL) IVORYERROR(err_msg, ierIvoryPaletteNotAssigned);
   return TheIvoryDraw->Palette->Color[palindex];
}
//---------------------------------------------------------------------------
__fastcall TIvoryDraw::TIvoryDraw(Classes::TComponent* AOwner)
 : TComponent(AOwner),
   FForm(NULL), FDirectDraw(NULL),
   FPrimarySurface(NULL), FBackSurface(NULL), FInternalBackSurface(NULL),
   FUserBackSurface(NULL), FPaintingStarted(NULL), FPalette(NULL),
   FAutoActivate(true), FAutoHideCursor(true), FExclusiveMode(false),
   FVideoMode(ivm640x480), FWidth(640), FHeight(480),
   FActive(false),
   FOriginalFormCreate(NULL), FOnCreate(NULL), FOnDestroy(NULL),
   FOnActivate(NULL), FOnDeactivate(NULL), FOnSurfaceLost(NULL),
   FOnException(NULL)
{
   const char* ivoryfuncname = "TIvoryDraw";
   if(hDll == NULL)
   // this is the first time that you dropped a component on the form
   {
      hDll = LoadLibrary("DDRAW.DLL");
      if(hDll == NULL) IVORYERROR("Couldn\'t load DDRAW.DLL", ierUnknown);
      PDirectDrawCreate = (PROCDirectDrawCreate)GetProcAddress(hDll, "DirectDrawCreate");
      if(PDirectDrawCreate == NULL) IVORYERROR("Couldn\'t initialize DDRAW.DLL", ierUnknown);
   }
   if(dynamic_cast<TForm*>(AOwner))
      FForm = (TForm*)AOwner;
   else
      IVORYERROR("Please drop TIvoryDraw immediately on a TForm object", ierOther);
   FHandle = Form->Handle;
   VideoMode = ivm640x480;
   if(!ComponentState.Contains(csDesigning))
      Application->OnException = HandleException;
   if(TheIvoryDraw == NULL) TheIvoryDraw = this;

   if(ComponentState.Contains(csDesigning))
      Form->BorderStyle = bsNone;
   else
   {
      if(Form->BorderStyle != bsNone && Form->BorderStyle != bsSizeToolWin && Form->BorderStyle != bsToolWindow)
      // hide the caption if necessary
      {
         unsigned curr = GetWindowLong(Form->Handle, GWL_STYLE);
         unsigned currex = GetWindowLong(Form->Handle, GWL_EXSTYLE);
         if((curr & WS_CAPTION) == WS_CAPTION)
         {
            if(Form->BorderStyle == bsSingle || Form->BorderStyle == bsSizeable)
               SetWindowLong(Form->Handle, GWL_STYLE, (curr & (!WS_CAPTION)) | WS_BORDER);
            else if(Form->BorderStyle == bsDialog)
            {
               SetWindowLong(Form->Handle, GWL_EXSTYLE, (currex & (!WS_EX_DLGMODALFRAME)));
               SetWindowLong(Form->Handle, GWL_STYLE, (curr & (!WS_CAPTION) & (!WS_POPUP)) | WS_BORDER);
            }
            Form->Height = Form->Height + GetSystemMetrics(SM_CYCAPTION);
         }
      }

      FOriginalFormCreate = Form->OnCreate;
      Form->OnCreate = FormCreate;
   }
   FPrimarySurface = new TIvorySurface(this);
   FPrimarySurface->FSurfaceType = stPrimary;
   FInternalBackSurface = FBackSurface = new TIvorySurface(this);
   FBackSurface->FSurfaceType = stBack;
}
//---------------------------------------------------------------------------
__fastcall TIvoryDraw::~TIvoryDraw()
{
   Active = false;         // Back to normal Windows screen mode
   if(Application->OnException == HandleException)
      Application->OnException = NULL;
   if(hDll != NULL) FreeLibrary(hDll);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::SetActive(bool active)
{
   if(active != Active)
   {
      if(active)
         Activate();
      else
         Deactivate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::SetExclusiveMode(bool exclusive)
{
   if(!Active)
      FExclusiveMode = exclusive;
   // You can only set ExclusiveMode if Ivory Draw is not activated yet.
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::SetPalette(TIvoryPalette* palette)
{
   FPalette = palette;
   if(Active && palette) // If Ivory Draw is active and palette is not NULL, use the palette
      RefreshPalette();
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::SetUserBackSurface(TCustomIvorySurface* surface)
{
   if(!Active)
   {
      FUserBackSurface = surface;
      if(surface)
         FBackSurface = surface;
      else
         FBackSurface = FInternalBackSurface;
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::SetVideoMode(TIvoryVideoMode ivm)
{
   if(!Active)
   {
      FVideoMode = ivm;
      switch(ivm)
      {
         case ivm640x480:
            FWidth = 640; FHeight = 480; break;
         case ivm800x600:
            FWidth = 800; FHeight = 600; break;
         case ivm1024x768:
            FWidth = 1024; FHeight = 768; break;
         default:
            FWidth = 640; FHeight = 480; FVideoMode = ivm640x480; break;
      }
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::StartPainting(TCustomIvorySurface* Surface)
{
   if(FPaintingStarted == NULL)
      FPaintingStarted = Surface;
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::FinishPainting(TCustomIvorySurface* Surface)
{
   if(FPaintingStarted == Surface) // only the starter can finish
   {
      FPaintingStarted = NULL;
      for(int i = 0; i < DDSList->Count; i++)
         ((TCustomIvorySurface*)DDSList->Items[i])->FPainted = false;
      PrimarySurface->FPainted = false;
      BackSurface->FPainted = false; 
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::AddSurface(TCustomIvorySurface* surface)
{
   if(Active)
      DDSList->Add(surface);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::ReleaseSurface(TCustomIvorySurface* surface)
{
   if(Active)
   {
      for(int i = 0; i < DDSList->Count; i++)
      {
         if(DDSList->Items[i] == (void*)surface)
         {
            DDSList->Delete(i);
            break;
         }
      }
      surface->Release();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::AddPalette(TIvoryPalette* palette)
{
   if(Active)
      DDPList->Add(palette);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::ReleasePalette(TIvoryPalette* palette)
{
   if(Active)
   {
      for(int i = 0; i < DDPList->Count; i++)
      {
         if(DDPList->Items[i] == (void*)palette)
         {
            DDPList->Delete(i);
            break;
         }
      }
      palette->Release();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::AutoCreateSurfaces()
{
   for(int i = 0; i < Form->ComponentCount; i++)
   {
      if(dynamic_cast<TCustomIvorySurface*>(Form->Components[i]))
      {
         TCustomIvorySurface* surface = (TCustomIvorySurface*)(Form->Components[i]);
         if(surface->AutoCreate && surface->DirectDrawSurface == NULL)
            surface->Create();
      }
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::RestoreSurfacesIfLost()
{
   TCustomIvorySurface* s;
   bool sl = false;
   s = PrimarySurface;
   if(s->DirectDrawSurface->IsLost() != DD_OK)
   {
      s->/*DirectDrawSurface->*/Restore();
      s->Valid = false;
      if(s->OnSurfaceLost) s->OnSurfaceLost(s);
      sl = true;
   }
   s = BackSurface;
   if(s->DirectDrawSurface->IsLost() != DD_OK)
   {
      s->/*DirectDrawSurface->*/Restore();
      if(s->OnSurfaceLost) s->OnSurfaceLost(s);
      sl = true;
   }
   for(int i = 0; i < DDSList->Count; i++)
   {
      s = (TCustomIvorySurface*)DDSList->Items[i];
      s->Valid = false;
      if(s != BackSurface && s->DirectDrawSurface->IsLost() != DD_OK)
      {
         s->/*DirectDrawSurface->*/Restore();
         if(s->OnSurfaceLost) s->OnSurfaceLost(s);
         sl = true;
      }
   }
   if(sl)
   {
      if(OnSurfaceLost) OnSurfaceLost(this);
      InvalidateSurfaces();
      BackSurface->Paint(true);
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::Activate()
{
   const char* cooperative_err_msg = "Couldn\'t set cooperative level";
   const char* ivoryfuncname = "Activate";
   Form->OnPaint = FormPaint;
//Form->Cursor = crNone;

   DDSList = new TList();
   DDPList = new TList();

   if(TheIvoryDraw && TheIvoryDraw != this) TheIvoryDraw->Active = false;
      // Only 1 Ivory Draw can be active at the same time; we must
      // deactivate the other one before activiting this one

   TheIvoryDraw = this;

   if(TheDirectDraw == NULL)
   {
      HRESULT ddrval = PDirectDrawCreate(NULL, &TheDirectDraw, NULL);
      if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t create IDirectDraw object", ierDirectX, ddrval);
   }
   FDirectDraw = TheDirectDraw;
   if(ExclusiveMode) // if exclusive mode is enabled, change to exclusive mode
   {
      HRESULT ddrval;
      ddrval = DirectDraw->SetCooperativeLevel(Handle, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT);
      if(ddrval != DD_OK) IVORYDDERROR(cooperative_err_msg, ierDirectX, ddrval);
      ddrval = DirectDraw->SetDisplayMode(Width, Height, 8);
      if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t set display mode", ierDirectX, ddrval);
   }
   else
   {
      if(Form)
      {
         Form->Left = 0;
         Form->Top = 0;
         Form->ClientWidth = Width;
         Form->ClientHeight = Height;
      }
      HRESULT ddrval = DirectDraw->SetCooperativeLevel(Handle, DDSCL_NORMAL);
      if(ddrval != DD_OK) IVORYDDERROR(cooperative_err_msg, ierDirectX, ddrval);
   }
   FActive = true;
   // this is the place where we are setting the Active bit, because now
   // the DirectDraw object is created and the cooperative mode is set,
   // so any DirectDraw function can be called; note that the primary and the
   // back surfaces are not set up yet

   if(FPrimarySurface->DirectDrawSurface == NULL) // If no primary surface is created, create it
      FPrimarySurface->CreatePrimary();
//   if(FUserBackSurface)
//      FBackSurface
   if(FBackSurface->DirectDrawSurface == NULL) // If no back surface is created, create it
      FBackSurface->CreateBack(FPrimarySurface);
   if(Palette) // If Palette is specified, use it
   {
      TIvoryPalette* ip = Palette;
      Palette = NULL;
      Palette = ip;
   }
   AutoCreateSurfaces();
//   BackSurface->Paint();

   if(OnActivate) OnActivate(this);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::Deactivate()
{
   if(OnDeactivate) OnDeactivate(this);

   DirectDraw->FlipToGDISurface();
   DirectDraw->SetCooperativeLevel(Handle, DDSCL_NORMAL);
      // Back to normal Windows screen mode

   FActive = false;
   PrimarySurface->Release();        // Release the primary surface
   FInternalBackSurface->Release();  // Release the back surface
   while(DDSList->Count)
   {
      ((TCustomIvorySurface*)DDSList->Items[0])->Release();
      DDSList->Delete(0);
   }
   delete DDSList;
   while(DDPList->Count)
   {
      ((TIvoryPalette*)DDPList->Items[0])->Release();
      DDPList->Delete(0);
   }
   delete DDPList;
//   if(Palette)
//      Palette->Release();  // Release the current palette (if there is one)
   if(DirectDraw)
      DirectDraw->Release();
   TheDirectDraw = NULL;
   FDirectDraw = NULL;
   TheIvoryDraw = NULL;
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::RefreshPalette()
{
   const char* ivoryfuncname = "RefreshPalette";
   if(Active && Palette)
   {
      HRESULT ddrval;
      if(Palette->DirectDrawPalette == NULL) // if the palette is not created yet, create it
         Palette->Create();
      ddrval = PrimarySurface->DirectDrawSurface->SetPalette(Palette->DirectDrawPalette);
      if(ExclusiveMode && ddrval != DD_OK) IVORYDDERROR("Couldn\'t set palette", ierDirectX, ddrval);
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::AfterConstruction()
{
   if(OnCreate) OnCreate(this);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::BeforeDestruction()
{
   if(OnDestroy) OnDestroy(this);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::FormCreate(TObject* Sender)
{
   if(FOriginalFormCreate) FOriginalFormCreate(Sender);
   HandleFormCreate(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::FormPaint(TObject* Sender)
{
   HandleFormPaint(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::Flip()
{
   if(!Active) return;
   if(ExclusiveMode)
   {
      while( 1 )
      {
         HRESULT ddrval;
         ddrval = PrimarySurface->DirectDrawSurface->Flip( NULL, 0 );
//       Application->ProcessMessages();
         if(ddrval == DD_OK)
            break;
         if(ddrval == DDERR_SURFACELOST)
         {
            PrimarySurface->Valid = false;
            ddrval = PrimarySurface->DirectDrawSurface->Restore();
            if(ddrval == DD_OK)
               break;
         }
         if(ddrval != DDERR_WASSTILLDRAWING)
            break;
      }
   }
   else
   {
      RECT r1, r;
      Application->ProcessMessages();
      GetWindowRect(Handle, &r);
      r1 = Rect(0, 0, Width, Height);
      Application->ProcessMessages();
      PrimarySurface->DirectDrawSurface->Blt(&r, BackSurface->DirectDrawSurface, &r1, 0, NULL);
      Application->ProcessMessages();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::HandleFormCreate(TObject* Sender)
{
   if(FAutoHideCursor) ((TForm*)Sender)->Cursor = crNone;
   if(FAutoActivate) Active = true;
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::HandleFormPaint(TObject* Sender)
{
   RestoreSurfacesIfLost();
   /*if(ExclusiveMode)*/ BackSurface->Paint(true);
   Flip();
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::HandleException(TObject* Sender, Exception* E)
{
   if(OnException) OnException(Sender, E);
   else
   {
      EIvoryError* err;
      try
      {
         TForm* Form = dynamic_cast<TForm*>(Owner);
         if(Form)
         {
            Form->OnActivate = NULL;
            Form->OnClick = NULL;
            Form->OnClose = NULL;
            Form->OnCloseQuery = NULL;
            Form->OnDblClick = NULL;
            Form->OnDeactivate = NULL;
            Form->OnHide = NULL;
            Form->OnKeyDown = NULL;
            Form->OnKeyPress = NULL;
            Form->OnKeyUp = NULL;
            Form->OnMouseMove = NULL;
            Form->OnMouseDown = NULL;
            Form->OnMouseUp = NULL;
            Form->OnPaint = NULL;
            Form->OnResize = NULL;
         }
         err = dynamic_cast<EIvoryError*>(E);
         if(TheIvoryDraw) TheIvoryDraw->Active = false;
      }
      catch(...)
      {
      }
      try
      {
         if(err)
         {
            if(err->Reason == ierDirectX)
               ShowMessage("Error: " + err->Message + "\nClass: " + err->ClassString + "\nFunction: " + err->Function + "\nReason: " + err->ReasonString + "\nDDResult: " + err->DDResultString);
            else
               ShowMessage("Error: " + err->Message + "\nClass: " + err->ClassString + "\nFunction: " + err->Function + "\nReason: " + err->ReasonString);
         }
         else
            ShowMessage(E->Message);
      }
      catch(...)
      {
      }
//      if(MessageBox(NULL, "Press Ok to shut down the program", "Shut down", MB_OK)==IDOK)
      Application->Terminate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::InvalidateSurfaces()
{
   for(int i = 0; i < DDSList->Count; i++)
      ((TCustomIvorySurface*)DDSList->Items[i])->Valid = false;
   BackSurface->Valid = false;
}
//---------------------------------------------------------------------------

