//*****************************************************
//**   Ivory Draw 1.0 Beta 4                         **
//**   DirectDraw component package for C++Builder   **
//*****************************************************

// (C) 1998 Tamas Demjen
// email: demjen@usa.net
// Web: http://members.xoom.com/demjen
// Ivory Draw on the Web: http://members.xoom.com/demjen/bcb/ivory.html

// This product is freeware, as described in license.txt
//---------------------------------------------------------------------------
#ifndef IvoryDrawH
#define IvoryDrawH

#include <vcl\SysUtils.hpp>
#include <vcl\Controls.hpp>
#include <vcl\Classes.hpp>
#include <vcl\Forms.hpp>
#include <ddraw.h>
#include "IvoryError.h"
#include "IvoryPalette.h"
#include "CustomIvorySurface.h"
//---------------------------------------------------------------------------
enum TIvoryVideoMode { ivm640x480, ivm800x600, ivm1024x768 };

class IVORY_package TIvoryDraw : public TComponent
{
   friend TCustomIvorySurface;
   friend TIvoryPalette;
private:
// internal handles
   HWND FHandle;
   TForm* FForm; // parent form (the parent must be a form!)
   LPDIRECTDRAW FDirectDraw; // DirectX object IDirectDraw
// child objects
   TCustomIvorySurface* FPrimarySurface; // the primary surface
   TCustomIvorySurface* FBackSurface; // the current back surface
   TCustomIvorySurface* FInternalBackSurface; // the internal back surface
   TCustomIvorySurface* FUserBackSurface; // the user back surface
   TIvoryPalette* FPalette; // the palette
   TCustomIvorySurface* FPaintingStarted; // the object that started the painting
// initial settings
   bool FAutoActivate;
   bool FAutoHideCursor;
   bool FExclusiveMode;
   TIvoryVideoMode FVideoMode;
   int FWidth;
   int FHeight;
// current state
   bool FActive;
// internal messages
   Classes::TNotifyEvent FOriginalFormCreate;
// public messages
   Classes::TNotifyEvent FOnCreate;
   Classes::TNotifyEvent FOnDestroy;
   Classes::TNotifyEvent FOnActivate;
   Classes::TNotifyEvent FOnDeactivate;
   Classes::TNotifyEvent FOnSurfaceLost;
   TExceptionEvent FOnException;
// getter and setter functions
   void __fastcall SetActive(bool active);
   void __fastcall SetExclusiveMode(bool exclusive);
   void __fastcall SetPalette(TIvoryPalette* palette);
   void __fastcall SetUserBackSurface(TCustomIvorySurface* surface);
   void __fastcall SetVideoMode(TIvoryVideoMode ivm);
// private functions
   void __fastcall StartPainting(TCustomIvorySurface* Surface);
   void __fastcall FinishPainting(TCustomIvorySurface* Surface);
   void __fastcall AddSurface(TCustomIvorySurface* surface);
   void __fastcall ReleaseSurface(TCustomIvorySurface* surface);
   void __fastcall AddPalette(TIvoryPalette* palette);
   void __fastcall ReleasePalette(TIvoryPalette* palette);
// internal message handling
   void __fastcall FormCreate(TObject* Sender);
   void __fastcall FormPaint(TObject* Sender);
protected:
// internal functions
   void __fastcall AutoCreateSurfaces();
   void __fastcall RestoreSurfacesIfLost();
   void __fastcall Activate();
   void __fastcall Deactivate();
   void __fastcall RefreshPalette();
// internal message handling
   virtual void __fastcall AfterConstruction(); // override TCustomForm::AfterConstruction
   virtual void __fastcall BeforeDestruction(); // override TCustomForm::BeforeDestruction
// internal properties
//   __property TCustomIvorySurface* PaintingStarted = { read=FPaintingStarted, write=FPaintingStarted };
public:
// constructor
   __fastcall TIvoryDraw(TComponent* AOwner);
// destructor
   virtual __fastcall ~TIvoryDraw();
// public functions
   void __fastcall Flip();
   void __fastcall HandleFormCreate(TObject* Sender);
   void __fastcall HandleFormPaint(TObject* Sender);
   void __fastcall HandleException(TObject* Sender, Exception* E);
   void __fastcall InvalidateSurfaces();
// public properties
   __property HWND Handle = { read=FHandle };
   __property TForm* Form = { read=FForm };
   __property LPDIRECTDRAW DirectDraw = { read=FDirectDraw };
   __property TCustomIvorySurface* PrimarySurface = { read=FPrimarySurface };
   __property TCustomIvorySurface* BackSurface = { read=FBackSurface };
   __property int Width = { read=FWidth };
   __property int Height = { read=FHeight };
   __property bool Active = { read=FActive, write=SetActive };
__published:
// published properties
   __property TCustomIvorySurface* UserBackSurface = { read=FUserBackSurface, write=SetUserBackSurface, default=NULL };
   __property TIvoryPalette* Palette = { read=FPalette, write=SetPalette, default=NULL};
   __property bool AutoActivate = { read=FAutoActivate, write=FAutoActivate, default=true };
   __property bool AutoHideCursor = { read=FAutoHideCursor, write=FAutoHideCursor, default=true };
   __property bool ExclusiveMode = { read=FExclusiveMode, write=SetExclusiveMode, default=false };
   __property TIvoryVideoMode VideoMode = { read=FVideoMode, write=SetVideoMode, default=ivm640x480 };
// published events
   __property Classes::TNotifyEvent OnCreate = { read=FOnCreate, write=FOnCreate, default=NULL };
   __property Classes::TNotifyEvent OnDestroy = { read=FOnDestroy, write=FOnDestroy, default=NULL };
   __property Classes::TNotifyEvent OnActivate = { read=FOnActivate, write=FOnActivate, default=NULL };
   __property Classes::TNotifyEvent OnDeactivate = { read=FOnDeactivate, write=FOnDeactivate, default=NULL };
   __property Classes::TNotifyEvent OnSurfaceLost = { read=FOnSurfaceLost, write=FOnSurfaceLost, default=NULL };
   __property TExceptionEvent OnException = { read=FOnException, write=FOnException, default=NULL };
};

extern TIvoryDraw* TheIvoryDraw;
extern TColor __fastcall ToTColor(int palindex);
//---------------------------------------------------------------------------
#endif
