//*****************************************************
//**   Ivory Draw 1.0 Beta 4                         **
//**   DirectDraw component package for C++Builder   **
//*****************************************************

// (C) 1998 Tamas Demjen
// email: demjen@usa.net
// Web: http://members.xoom.com/demjen
// Ivory Draw on the Web: http://members.xoom.com/demjen/bcb/ivory.html

// This product is freeware, as described in license.txt
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "IvorySurface.h"
#include "IvoryDraw.h"
//---------------------------------------------------------------------------
#ifdef __IvoryBCB3__
#pragma package(smart_init)
#endif
#pragma link "IvoryCanvas"
#pragma link "IvoryError"
//---------------------------------------------------------------------------
// internal variables:
static const TIvoryErrorClass iec = iecIvorySurface;
//---------------------------------------------------------------------------
static inline void ValidCtrCheckSurface(TIvorySurface*)
{
   new TIvorySurface(NULL);
}
//---------------------------------------------------------------------------
__fastcall TIvorySurface::TIvorySurface(TComponent* AOwner)
 : TCustomIvorySurface(AOwner),
   FMapper(NULL), FBackgroundBitmap(NULL), FBackgroundSurface(NULL),
   FMapLeft(0), FMapTop(0), FMapColumns(0), FMapRows(0), FMapEnabled(false),
   FMap(NULL), FDrawBackground(false), FBackgroundDrawStyle(idsStretch),
   FOnGetMap(NULL), FAfterDrawMap(NULL)
{
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetMapper(TIvoryMapper* mapper)
{
   if(FMapper != mapper)
   {
      FMapper = mapper;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetMapLeft(int ml)
{
   if(FMapLeft != ml)
   {
      FMapLeft = ml;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetMapTop(int mt)
{
   if(FMapTop != mt)
   {
      FMapTop = mt;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetMapColumns(int mc)
{
   if(FMapColumns != mc)
   {
      FMapColumns = mc;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetMapRows(int mr)
{
   if(FMapRows != mr)
   {
      FMapRows = mr;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetMapEnabled(bool me)
{
   if(FMapEnabled != me)
   {
      FMapEnabled = me;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetDrawBackground(bool db)
{
   if(FDrawBackground != db)
   {
      FDrawBackground = db;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetBackgroundDrawStyle(TIvoryDrawStyle bds)
{
   if(FBackgroundDrawStyle != bds)
   {
      FBackgroundDrawStyle = bds;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetBackgroundBitmap(TIvoryBitmap* bitmap)
{
   if(FBackgroundBitmap != bitmap)
   {
      FBackgroundBitmap = bitmap;
      FBackgroundSurface = NULL;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetBackgroundSurface(TCustomIvorySurface* surface)
{
   if(FBackgroundSurface != surface)
   {
      FBackgroundBitmap = NULL;
      FBackgroundSurface = surface;
      Invalidate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::RealizePaint()
{
   if((ClearBackground && !DrawBackground) ||
      (ClearBackground && DrawBackground && BackgroundDrawStyle == idsCenter))
   {
      if(TheIvoryDraw->ExclusiveMode)
         FillRect(Rect(0, 0, Width, Height), BackgroundColor);
      else
      {
         GrabCanvas();
         TColor c = ToTColor(BackgroundColor);
         Canvas->Pen->Color = c;
         Canvas->Brush->Color = c;
         Canvas->Rectangle(0, 0, Width, Height);
      }
   }
   if(DrawBackground)
   {
      if(BackgroundBitmap)
      {
         if(Canvas->Handle == NULL) GrabCanvas();
         if(BackgroundDrawStyle == idsStretch)
            Canvas->StretchDraw(Rect(0, 0, Width, Height), BackgroundBitmap);
         else if(BackgroundDrawStyle == idsCenter)
         // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
         // !!!!here we should check if the bitmap is smaller than the surface...!!!!
            Canvas->Draw((Width - BackgroundBitmap->Width) >> 2, (Height - BackgroundBitmap->Height) >> 2, BackgroundBitmap);
         else if(BackgroundDrawStyle == idsTile)
            Canvas->TileDraw(Rect(0, 0, Width, Height), BackgroundBitmap);
      }
      else if(BackgroundSurface)
      {
         PaintSubSurface(BackgroundSurface);
         if(BackgroundDrawStyle == idsStretch)
            StretchCopySurface(Rect(0, 0, Width, Height), BackgroundSurface);
         else if(BackgroundDrawStyle == idsCenter)
         // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
         // !!!!here we should check if the background surface is smaller than the surface...!!!!
            CopySurface((Width - BackgroundSurface->Width) >> 2, (Height - BackgroundSurface->Height) >> 2, BackgroundSurface);
         else if(BackgroundDrawStyle == idsTile)
            TileCopySurface(Rect(0, 0, Width, Height), BackgroundSurface);
      }
   }
   if(OnPaint) OnPaint(this);
   if(MapEnabled && Mapper) DrawMap();
   if(AfterDrawMap) AfterDrawMap(this);
   if(Canvas->Handle) ReleaseCanvas();
   if(Locked) Unlock();
}
//---------------------------------------------------------------------------
// draw a map on the surface. A mapper must be set and the OnGetMap event must be written
void __fastcall TIvorySurface::DrawMap()
{
   if(MapEnabled && Mapper && MapRows && MapColumns && (OnGetMap || Map))
   {
      if(Mapper->Bitmap && Canvas->Handle == NULL) GrabCanvas();
      else if(Mapper->BitmapSurface)
      {
         if(Canvas->Handle) ReleaseCanvas();
         if(Locked) Unlock();
      }
      int maxindex = min(Mapper->TileColumns * Mapper->TileRows, Mapper->Tiles);
      int index = 0;
      int x, y;
      int row, col;
      if(Mapper->Bitmap)
      {
         if(Map) // if Map is specified, use it
         {
            int* map = Map;
            for(row = 0, y = MapTop; row < MapRows; row++, y += Mapper->TileHeight)
            {
               for(col = 0, x = MapLeft; col < MapColumns; col++, x += Mapper->TileWidth)
               {
                  index = *map++;
                  if(index >= 0 && index < maxindex)
                     Canvas->MapDraw(x, y, Mapper->Bitmap->Handle, Mapper->TileRect[index]);
               }
            }
         }
         else // if Map is NULL, use OnGetMap
         {
            for(row = 0, y = MapTop; row < MapRows; row++, y += Mapper->TileHeight)
            {
               for(col = 0, x = MapLeft; col < MapColumns; col++, x += Mapper->TileWidth)
               {
                  index = -1;
                  OnGetMap(this, row, col, index);
                  if(index >= 0 && index < maxindex)
                     Canvas->MapDraw(x, y, Mapper->Bitmap->Handle, Mapper->TileRect[index]);
               }
            }
         }
      }
      else if(Mapper->BitmapSurface)
      {
         PaintSubSurface(Mapper->BitmapSurface);
         if(Map) // if Map is specified, use it
         {
            int* map = Map;
            for(row = 0, y = MapTop; row < MapRows; row++, y += Mapper->TileHeight)
            {
               for(col = 0, x = MapLeft; col < MapColumns; col++, x += Mapper->TileWidth)
               {
                  index = *map++;
                  if(index >= 0 && index < maxindex)
                     CopySurface(x, y, Mapper->BitmapSurface, Mapper->TileRect[index]);
               }
            }
         }
         else // if Map is NULL, use OnGetMap
         {
            for(row = 0, y = MapTop; row < MapRows; row++, y += Mapper->TileHeight)
            {
               for(col = 0, x = MapLeft; col < MapColumns; col++, x += Mapper->TileWidth)
               {
                  index = -1;
                  OnGetMap(this, row, col, index);
                  if(index >= 0 && index < maxindex)
                     CopySurface(x, y, Mapper->BitmapSurface, Mapper->TileRect[index]);
               }
            }
         }
      }
   }
}
//---------------------------------------------------------------------------

